import psycopg2
from psycopg2.extensions import AsIs
from pg4nosql import DEFAULT_POSTGRES_DATABASE, DEFAULT_POSTGRES_PORT, DEFAULT_POSTGRES_HOST
from pg4nosql.PostgresNoSQLDatabase import PostgresNoSQLDatabase


class PostgresNoSQLClient(object):
    """
    Creates new connections and
    """

    __SQL_DATABASE_EXISTS = 'SELECT EXISTS(SELECT datname FROM pg_database WHERE datname=%s)'
    __SQL_CREATE_DATABASE = 'CREATE DATABASE %s'
    __SQL_DROP_DATABASE = 'DROP DATABASE IF EXISTS %s'

    def __init__(self, host=DEFAULT_POSTGRES_HOST, database=DEFAULT_POSTGRES_DATABASE, port=DEFAULT_POSTGRES_PORT,
                 user=None, password=None):
        # public fields
        self.host = host
        self.database = database
        self.port = port

        # private
        self.__user = user
        self.__password = password
        self.__connection = None
        self.__cursor = None

        # helps to make multiple functions with one connection
        self.__connect_counter = 0

    def __commit(self):
        self.__connection.commit()

    def __connect(self):
        self.__connect_counter += 1
        if not self.__connection:
            self.__connection = psycopg2.connect(host=self.host, database=self.database, port=self.port,
                                                 user=self.__user, password=self.__password)
            self.__connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            self.__cursor = self.__connection.cursor()

    def __close(self):
        self.__connect_counter -= 1

        if self.__connect_counter is 0:
            self.__connection.close()
            self.__connection = None

        return self.__connect_counter is 0

    def __create_db_object(self, database_name):
        return PostgresNoSQLDatabase(name=database_name, host=self.host, port=self.port, user=self.__user,
                                     password=self.__password)

    def get_database(self, database_name):
        self.__connect()
        db_object = None

        if self.database_exists(database_name):
            db_object = self.__create_db_object(database_name)

        self.__close()
        return db_object

    def get_or_create_database(self, database_name):
        self.__connect()
        db = self.get_database(database_name)
        if not db:
            db = self.create_database(database_name)
        self.__close()
        return db

    def create_database(self, database_name):
        self.__connect()
        self.__cursor.execute(self.__SQL_CREATE_DATABASE, (AsIs(database_name),))
        self.__commit()
        self.__close()
        return self.__create_db_object(database_name)

    def drop_database(self, database_name):
        self.__connect()
        self.__cursor.execute(self.__SQL_DROP_DATABASE, (AsIs(database_name),))
        self.__close()

    def database_exists(self, database_name):
        self.__connect()
        exists = False
        try:
            self.__cursor.execute(self.__SQL_DATABASE_EXISTS, (database_name,))
            exists = self.__cursor.fetchone()[0]
        except psycopg2.Error as e:
            print e
        self.__close()
        return exists

    def __getitem__(self, item):
        return self.get_or_create_database(item)



import json
from psycopg2.extensions import AsIs
from psycopg2.extras import RealDictCursor
from pg4nosql import DEFAULT_JSON_COLUMN_NAME, DEFAULT_ROW_IDENTIFIER
from pg4nosql.PostgresNoSQLResultItem import PostgresNoSQLResultItem


class PostgresNoSQLTable(object):
    """
    Represents the a postgres table.
    """

    __SQL_INSERT_JSON = "INSERT INTO %s(" + DEFAULT_JSON_COLUMN_NAME + " %s) VALUES(%s %s) RETURNING " + DEFAULT_ROW_IDENTIFIER
    __SQL_QUERY_JSON = 'SELECT %s FROM %s WHERE %s'
    __SQL_GET_JSON = 'SELECT * FROM %s WHERE ' + DEFAULT_ROW_IDENTIFIER + '=%s'
    __SQL_GET_COLUMNS = 'select column_name from information_schema.columns where table_name = %s'
    __SQL_DELETE_JSON = 'DELETE FROM %s WHERE ' + DEFAULT_ROW_IDENTIFIER + '=%s'
    __SQL_UPDATE_JSON = 'UPDATE %s SET ' + DEFAULT_JSON_COLUMN_NAME + '=%s %s WHERE ' + DEFAULT_ROW_IDENTIFIER + '=%s;'

    def __init__(self, name, connection):
        self.name = name
        self.connection = connection
        self.connection.cursor_factory = RealDictCursor
        self.cursor = self.connection.cursor()

    @staticmethod
    def __to_sql_string(obj):
        if obj is None:
            return AsIs(obj)
        return str(obj)

    @staticmethod
    def __to_nullable_string(obj):
        if obj is None:
            return 'Null'
        return "'" + str(obj) + "'"

    def commit(self):
        self.connection.commit()

    def put(self, json_data, **relational_data):
        # filter none values out
        relational_data = {k: v for (k, v) in relational_data.items() if v is not None}

        # todo: replace string concatenation with a beautiful solution
        relational_data_columns = ''
        relational_data_values = ''

        if relational_data:
            relational_data_columns = ',' + ",".join(relational_data.keys())
            relational_data_values = ",'" + "','".join(map(str, relational_data.values())) + "'"

        self.cursor.execute(self.__SQL_INSERT_JSON, (AsIs(self.name),
                                                     AsIs(relational_data_columns), json.dumps(json_data),
                                                     AsIs(relational_data_values)))
        return self.cursor.fetchone()[DEFAULT_ROW_IDENTIFIER]

    def save(self, record):
        record = record.get_record()

        data = record.pop(DEFAULT_JSON_COLUMN_NAME)
        object_id = record.pop(DEFAULT_ROW_IDENTIFIER)

        relational_data_sql = ''.join(
            ", %s=%s" % (key, self.__to_nullable_string(val)) for (key, val) in record.iteritems())

        self.cursor.execute(self.__SQL_UPDATE_JSON, (AsIs(self.name),
                                                     json.dumps(data), AsIs(relational_data_sql), object_id))

    def get(self, object_id):
        self.cursor.execute(self.__SQL_GET_JSON, (AsIs(self.name), object_id))
        record = self.cursor.fetchone()

        if record is None:
            return record

        return PostgresNoSQLResultItem(record)

    def query(self, query='True', columns='*'):
        self.cursor.execute(self.__SQL_QUERY_JSON, (AsIs(columns), AsIs(self.name), AsIs(query)))
        rows = [item for item in self.cursor.fetchall()]
        items = map(lambda r: PostgresNoSQLResultItem(r), rows)
        return items

    def query_one(self, query='True', columns='*'):
        result = self.query(query, columns)
        if not result:
            return None
        return result[0]

    def get_columns(self):
        self.cursor.execute(self.__SQL_GET_COLUMNS, (self.name,))
        columns = map(lambda m: m['column_name'], self.cursor.fetchall())
        return columns

    def delete(self, object_id):
        self.cursor.execute(self.__SQL_DELETE_JSON, (AsIs(self.name), object_id))

    def execute(self, sql_query):
        self.cursor.execute(sql_query)
        return self.cursor.fetchall()



from pg4nosql import DEFAULT_JSON_COLUMN_NAME


class PostgresNoSQLResultItem(object):

    def __init__(self, inner_result):
        self.__inner_result = inner_result
        self.json = self.PostgresNoSQLJSONDocument(self.__inner_result)
        self.columns = self.__inner_result.keys()

    def __str__(self):
        return str(self.__inner_result)

    def __setitem__(self, key, value):
        self.__inner_result[key] = value

    def __getitem__(self, item):
        return self.__inner_result[item]

    def get_record(self):
        return self.__inner_result

    class PostgresNoSQLJSONDocument(object):
        def __init__(self, inner_result):
            self.__inner_result = inner_result

        def __setitem__(self, key, value):
            self.__inner_result[DEFAULT_JSON_COLUMN_NAME][key] = value

        def __getitem__(self, item):
            return self.__inner_result[DEFAULT_JSON_COLUMN_NAME][item]

        def __str__(self):
            return str(self.__inner_result[DEFAULT_JSON_COLUMN_NAME])



from distutils.core import setup

setup(
    name='pg4nosql',
    version='0.2.4',
    packages=['pg4nosql'],
    url='https://github.com/cansik/pg4nosql',
    license='MIT License',
    author='Florian Bruggisser',
    author_email='florian@nexpose.ch',
    description='A simple psycopg2 based wrapper for nosql like database interaction with python.',
    download_url='https://github.com/cansik/pg4nosql/tarball/0.2.4',
    keywords=['postgres', 'nosql', 'wrapper', 'database', 'json'],
    classifiers=[],
)



import psycopg2
from psycopg2.extensions import AsIs
from pg4nosql import DEFAULT_JSON_COLUMN_NAME
from pg4nosql.PostgresNoSQLTable import PostgresNoSQLTable


class PostgresNoSQLDatabase(object):
    __SQL_CREATE_JSON_TABLE = 'CREATE TABLE %s (id SERIAL PRIMARY KEY %s, ' + DEFAULT_JSON_COLUMN_NAME + ' JSON);'
    __SQL_DROP_JSON_TABLE = 'DROP TABLE IF EXISTS %s;'
    __SQL_TABLE_EXISTS = "SELECT EXISTS(SELECT relname FROM pg_class WHERE relname=%s)"

    def __init__(self, name, host, port, user, password):
        self.name = name
        self.connection = psycopg2.connect(host=host, database=name, port=port, user=user, password=password)
        self.cursor = self.connection.cursor()

    def close(self):
        self.connection.close()

    def commit(self):
        self.connection.commit()

    def create_table(self, table_name, **relational_columns):
        # create additional columns string
        columns_str = ''.join(', %s %s' % (key, val) for (key, val) in relational_columns.iteritems())
        self.cursor.execute(self.__SQL_CREATE_JSON_TABLE, (AsIs(table_name), AsIs(columns_str)))
        self.commit()
        return PostgresNoSQLTable(table_name, self.connection)

    def drop_table(self, table_name):
        self.cursor.execute(self.__SQL_DROP_JSON_TABLE, (AsIs(table_name),))
        self.commit()

    def get_table(self, table_name):
        if self.table_exists(table_name):
            return PostgresNoSQLTable(table_name, self.connection)
        else:
            return None

    def get_or_create_table(self, table_name, **relational_columns):
        table = self.get_table(table_name)
        if not table:
            table = self.create_table(table_name, **relational_columns)
        return table

    def table_exists(self, table_name):
        exists = False
        try:
            self.cursor.execute(self.__SQL_TABLE_EXISTS, (table_name,))
            exists = self.cursor.fetchone()[0]
        except psycopg2.Error as e:
            print e
        return exists

    def __getitem__(self, item):
        return self.get_or_create_table(item)
